// SPDX-License-Identifier: GPL-2.0+
/* NetworkManager Connection editor -- Connection editor for NetworkManager
 *
 * Dan Williams <dcbw@redhat.com>
 *
 * Copyright 2008 - 2014 Red Hat, Inc.
 */

#ifndef __PAGE_VPN_H__
#define __PAGE_VPN_H__

#include <glib.h>
#include <glib-object.h>

#include "ce-page.h"

#define CE_TYPE_PAGE_VPN            (ce_page_vpn_get_type ())
#define CE_PAGE_VPN(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), CE_TYPE_PAGE_VPN, CEPageVpn))
#define CE_PAGE_VPN_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), CE_TYPE_PAGE_VPN, CEPageVpnClass))
#define CE_IS_PAGE_VPN(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), CE_TYPE_PAGE_VPN))
#define CE_IS_PAGE_VPN_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), CE_TYPE_PAGE_VPN))
#define CE_PAGE_VPN_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), CE_TYPE_PAGE_VPN, CEPageVpnClass))

typedef struct {
	CEPage parent;
} CEPageVpn;

typedef struct {
	CEPageClass parent;
} CEPageVpnClass;

typedef struct {
	char *add_detail_key;
	char *add_detail_val;
} CEPageVpnDetailData;

GType ce_page_vpn_get_type (void);

CEPage *ce_page_vpn_new (NMConnectionEditor *editor,
                         NMConnection *connection,
                         GtkWindow *parent,
                         NMClient *client,
                         const char **out_secrets_setting_name,
                         GError **error);

gboolean ce_page_vpn_can_export (CEPageVpn *page);

void vpn_connection_new (FUNC_TAG_PAGE_NEW_CONNECTION_IMPL,
                         GtkWindow *parent,
                         const char *detail,
                         gpointer detail_data,
                         NMConnection *connection,
                         NMClient *client,
                         PageNewConnectionResultFunc result_func,
                         gpointer user_data);

#endif  /* __PAGE_VPN_H__ */
