// SPDX-License-Identifier: GPL-2.0+
/* NetworkManager Applet -- allow user control over networking
 *
 * Dan Williams <dcbw@redhat.com>
 *
 * Copyright 2007 - 2015 Red Hat, Inc.
 */

#ifndef __CE_UTILS_H__
#define __CE_UTILS_H__

void utils_fake_return_key (GdkEventKey *event);

#endif /* __CE_UTILS_H__ */
